
/* The hash functions used for saveing keys             */
/* One of key_length or key_length_offset must be given */
/* Key length of 0 isn't allowed                        */

#include "hist.h"

#define NO_RECORD	((uint) -1)
#define LOWFIND 1
#define LOWUSED 2
#define HIGHFIND 4
#define HIGHUSED 8


#define TRUE     1
#define FALSE    0




static uint hist_mask(uint histnr,uint buffmax,uint maxlength);
static void hist_movelink(HIST_LINK *array,uint pos,uint next_link,uint newlink);
static uint calc_histnr(const char *key,uint length);
static int  histcmp(HIST *hist,HIST_LINK *pos,const char *key,uint length);
static int  histcmp_n(HIST *hist,HIST_LINK *pos,const char *key,uint length);


char* get_key(const char *key,uint *length)
{
 *length=strlen(key);
 return (char*) key;
}

char* get_trunc_key(const char *key,uint *length)
{
 *length=strlen(key)-PID;
 return (char*) key;
}


char* get_trunc_key_double_int(const char *key,uint *length)
{
    *length=strlen(key)-2*FREQ_SPACE-2;
     return (char*) key;
}

char* get_trunc_key_simple_int(const char *key,uint *length)
{
    *length=strlen(key)-FREQ_SPACE-1;
     return (char*) key;
}

void free_record(void *record)
{
  free(record);
}



/********************************************************/
int init_dynamic_array(HIST_DYNAMIC_ARRAY *array, uint element_size,
			   uint init_alloc, uint alloc_increment)
{

  if (!alloc_increment)
  { 
    alloc_increment=max((MALLOC_SIZE)/element_size,16);
    if (init_alloc > 8 && alloc_increment > init_alloc * 2)
      alloc_increment=init_alloc*2;
  }

  if (!init_alloc)
    init_alloc=alloc_increment;
  array->elements=0;
  array->max_element=init_alloc;
  array->alloc_increment=alloc_increment;
  array->size_of_element=element_size;
  if (!(array->buffer=(char*) malloc(element_size*init_alloc)))
  {
    array->max_element=0;
    return(TRUE);
  }
  return(FALSE);
}

	/* Alloc room for one element */

char*  alloc_dynamic(HIST_DYNAMIC_ARRAY *array)
{
  if (array->elements == array->max_element)
  {
    char *new_ptr;
    if (!(new_ptr=(char*) realloc(array->buffer,(array->max_element+
				     array->alloc_increment)*
				     array->size_of_element)))
      return 0;
    array->buffer=new_ptr;
    array->max_element+=array->alloc_increment;
  }
  return array->buffer+(array->elements++ * array->size_of_element);
}

int insert_dynamic(HIST_DYNAMIC_ARRAY *array, char* element)
{
  char* buffer;
  if (array->elements == array->max_element)
  {						/* Call only when nessesary */
    if (!(buffer=  ((char*) alloc_dynamic(array))))
      return TRUE;
  }
  else
  {
    buffer=array->buffer+(array->elements * array->size_of_element);
    array->elements++;
  }
  memcpy(buffer,element,(size_t) array->size_of_element);
  return FALSE;
}




	/* remove last element from array and return it */

char *pop_dynamic(HIST_DYNAMIC_ARRAY *array)
{
  if (array->elements)
    return array->buffer+(--array->elements * array->size_of_element);
  return 0;
}


int set_dynamic(HIST_DYNAMIC_ARRAY *array, char* element, uint idx)
{
  if (idx >= array->elements)
  {
    if (idx >= array->max_element)
    {
      uint size;
      char *new_ptr;
      size=(idx+array->alloc_increment)/array->alloc_increment;
      size*= array->alloc_increment;
      if (!(new_ptr=(char*) realloc(array->buffer,size*
				       array->size_of_element)))
	return TRUE;
      array->buffer=new_ptr;
      array->max_element=size;
    }
    bzero((char*) (array->buffer+array->elements*array->size_of_element),
	  (idx - array->elements)*array->size_of_element);
    array->elements=idx+1;
  }
  memcpy(array->buffer+(idx * array->size_of_element),element,
	 (size_t) array->size_of_element);
  return FALSE;
}


void get_dynamic(HIST_DYNAMIC_ARRAY *array, char* element, uint idx)
{
  if (idx >= array->elements)
  {
    bzero(element,array->size_of_element);
    return;
  }
  memcpy(element,array->buffer+idx*array->size_of_element,
	 (size_t) array->size_of_element);
}


void delete_dynamic(HIST_DYNAMIC_ARRAY *array)
{
  if (array->buffer)
  {
    free(array->buffer);
    array->buffer=0;
    array->elements=array->max_element=0;
  }
}


void delete_dynamic_element(HIST_DYNAMIC_ARRAY *array, uint idx)
{
  char *ptr=array->buffer+array->size_of_element*idx;
  array->elements--;
  memmove(ptr,ptr+array->size_of_element,
	  (array->elements-idx)*array->size_of_element);
}


void freeze_size(HIST_DYNAMIC_ARRAY *array)
{
  uint elements=max(array->elements,1);

  if (array->buffer && array->max_element != elements)
  {
    array->buffer=(char*) realloc(array->buffer,
				     elements*array->size_of_element);
    array->max_element=elements;
  }
}




/*******************************************************/

char*   hist_element(HIST *hist,ulong index)
{
  if (index < hist->records)
    return array_dynamic_element(&hist->array,index,HIST_LINK*)->data;
  return 0;
}



int hist_init(HIST *hist,uint size,uint key_offset,uint key_length,
		  char *(*get_key)(const char *,uint*),
		  void (*free_element)(void*))
{

  if (init_dynamic_array(&hist->array,sizeof(HIST_LINK),size,0))
    return(1);
  hist->key_offset=key_offset;
  hist->key_length=key_length;
  hist->blength=1;
  hist->records=0;
  hist->current_record=NO_RECORD;			/* For the future */
  hist->get_key=get_key;
  hist->free=free_element;
  return(0);
}



void hist_free(HIST *hist)
{
  uint i,records;
  HIST_LINK *data=array_dynamic_element(&hist->array,0,HIST_LINK*);

  if (hist->free)
    for (i=0,records=hist->records ; i < records ; i++)
      (*hist->free)(data[i].data);
  delete_dynamic(&hist->array);
  hist->records=0;
}

	/* some helper functions */

inline char* hist_key(HIST *hist,const char *record,uint *length)
{
  if (hist->get_key)
    return (*hist->get_key)(record,length);
  *length=hist->key_length;
  return (char*) record+hist->key_offset;
}

	/* Calculate pos according to keys */

static uint hist_mask(uint histnr,uint buffmax,uint maxlength)
{
  if ((histnr & (buffmax-1)) < maxlength) return (histnr & (buffmax-1));
  return (histnr & ((buffmax >> 1) -1));
}

static uint hist_rec_mask(HIST *hist,HIST_LINK *pos,uint buffmax,uint maxlength)
{
  uint length;
  char *key=hist_key(hist,pos->data,&length);
  return hist_mask(calc_histnr(key,length),buffmax,maxlength);
}

	/* Calc histvalue for a key */

static uint calc_histnr(const char *key,uint length)
{
  register uint nr=1, nr2=3;
  int reclength;
  
  while (length--)
  {
    nr^= (((nr & 63)+nr2)*((uint) (uchar) *key++))+ (nr << 8);
    nr2+=3;
  }
  return((uint) nr);
}


inline uint rec_histnr(HIST *hist,const char *record)
{
  uint length;
  char *key=hist_key(hist,record,&length);
  return calc_histnr(key,length);
}


	/* Search after a record based on a key */
	/* Sets info->current_ptr to found record */

char* hist_search(HIST *hist,const char *key,uint length)
{
  HIST_LINK *pos;
  uint flag,index;


  flag=1;
  if (hist->records)
  {
    index=hist_mask(calc_histnr(key,length ? length : hist->key_length),
		    hist->blength,hist->records);
    do
    {
      pos= array_dynamic_element(&hist->array,index,HIST_LINK*);
      if (!histcmp(hist,pos,key,length))
      {
	hist->current_record= index;
	return (pos->data);
      }
      if (flag)
      {
	flag=0;					/* Reset flag */
	if (hist_rec_mask(hist,pos,hist->blength,hist->records) != index)
	  break;				/* Wrong link */
      }
    }
    while ((index=pos->next) != NO_RECORD);
  }
  hist->current_record= NO_RECORD;
  return(0);
}

        /* Search for a  record with identical key */
	/* It also get the frequency of the record  */

char* hist_search_frequency(HIST *hist,const char *key,uint length,uint*  frequency)
{
  HIST_LINK *pos;
  uint flag,index;


  flag=1;
  if (hist->records)
  {
    index=hist_mask(calc_histnr(key,length ? length : hist->key_length),
		    hist->blength,hist->records);
    do
    {
      pos= array_dynamic_element(&hist->array,index,HIST_LINK*);
      if (!histcmp(hist,pos,key,length))
      {
	hist->current_record= index;
        *frequency=pos->freq;
	return (pos->data);
      }
      if (flag)
      {
	flag=0;					/* Reset flag */
	if (hist_rec_mask(hist,pos,hist->blength,hist->records) != index)
	  break;				/* Wrong link */
      }
    }
    while ((index=pos->next) != NO_RECORD);
  }
  hist->current_record= NO_RECORD;
  *frequency=0;
  return(0);
}


char* hist_search_frequency_n(HIST *hist,const char *key,uint length,uint*  frequency)
{
  HIST_LINK *pos;
  uint flag,index;


  flag=1;
  if (hist->records)
  {
    index=hist_mask(calc_histnr(key,length ? length : hist->key_length),
		    hist->blength,hist->records);
    do
    {
      pos= array_dynamic_element(&hist->array,index,HIST_LINK*);
      if (!histcmp(hist,pos,key,length))
      {
	hist->current_record= index;
        *frequency=pos->freq;
	return (pos->data);
      }
      /* if (flag)
      {
	flag=0;					// Reset flag 
	if (hist_rec_mask(hist,pos,hist->blength,hist->records) != index)
	  break;				// Wrong link
      } */
    }
    while ((index=pos->next) != NO_RECORD);
  }
  hist->current_record= NO_RECORD;
  *frequency=0;
  return(0);
}
   


	/* Get next record with identical key */
	/* Can only be called if previous calls was hist_search */

char* hist_next(HIST *hist,const char *key,uint length)
{
  HIST_LINK *pos;
  uint index;

  if (index=hist->current_record != NO_RECORD)
  {
    HIST_LINK *data=array_dynamic_element(&hist->array,0,HIST_LINK*);
    for (index=data[hist->current_record].next; index != NO_RECORD ; index=pos->next)
    {
      pos=data+index;
      if (!histcmp(hist,pos,key,length))
      {
	hist->current_record= index;
	return pos->data;
      }
    }
    hist->current_record=NO_RECORD;
  }
  return 0;
}

        /* Get next record with identical key   and  gets the frequency of the next record  */
	/* Can only be called if previous calls was hist_search */
char* hist_next_frequency(HIST *hist,const char *key,uint length,uint* frequency)
{
  HIST_LINK *pos;
  uint index;

  if (index=hist->current_record != NO_RECORD)
  {
    HIST_LINK *data=array_dynamic_element(&hist->array,0,HIST_LINK*);
    for (index=data[hist->current_record].next; index != NO_RECORD ; index=pos->next)
    {
      pos=data+index;
      if (!histcmp(hist,pos,key,length))
      {
	hist->current_record= index;
        *frequency=(pos->freq);
	return pos->data;
      }
    }
    hist->current_record=NO_RECORD;
  }
  *frequency=0;
  return 0;
}




	/* Change link from pos to new_link */

static void hist_movelink(HIST_LINK *array,uint find,uint next_link,uint newlink)
{
  HIST_LINK *old_link;
  do
  {
    old_link=array+next_link;
  }
  while ((next_link=old_link->next) != find);
  old_link->next= newlink;
  return;
}

	/* Compare a key in a record to a hole key. Retrun 0 if identical */

static int histcmp_n(HIST *hist,HIST_LINK *pos,const char *key,uint length)
{
  uint rec_keylength;
  char *rec_key=hist_key(hist,pos->data,&rec_keylength);
  return (memcmp(rec_key,key,length));
}



	/* Compare a key in a record to  a sub  string to key . Retrun 0 if identical */

static int histcmp(HIST *hist,HIST_LINK *pos,const char *key,uint length)
{
  uint rec_keylength;
  char *rec_key=hist_key(hist,pos->data,&rec_keylength);
  return (length && length != rec_keylength) || memcmp(rec_key,key,rec_keylength);
}





	/* Incremente la frequence  de 1  lorsque  un enregistrement existe dans l'histogramme  
           ou insert un nouveau couple (record,1) si l'enregistrement n'existe pas dans l'histogramme  */
int hist_insert(HIST *hist,const char *record)
{ 
  HIST_LINK *pos,*data;
  uint hist_nr,index,length;
  char *key;
  int  flag;
  
  flag=1;
  if (hist->records)
  { key=hist_key(hist,record,&length);
    index=hist_mask(calc_histnr(key,length ? length : hist->key_length),
                                       hist->blength,hist->records);
    do
    {
      pos= array_dynamic_element(&hist->array,index,HIST_LINK*);
      if (!histcmp(hist,pos,key,length))
      {
	hist->current_record= index;
        pos->freq++;
	return (0);
      }
      if (flag)
      {
	flag=0;					/* Reset flag */
	if (hist_rec_mask(hist,pos,hist->blength,hist->records) != index)
	  break;				/* Wrong link */
      }
    }
    while ((index=pos->next) != NO_RECORD);
 
  }
  hist->current_record= NO_RECORD;
  return(hist_add_entry(hist,record));
   
}


         /* Ajoute 'frequency' a la frequence  lorsque  un enregistrement existe dans l'histogramme  
           ou insert un nouveau couple (record,frequency) si l'enregistrement record  n'existe pas dans l'histogramme  */
int hist_insert_frequency(HIST *hist,const char *record,uint frequency)
{ 
  HIST_LINK *pos,*data;
  uint hist_nr,index,length;
  char *key;
  int flag ;  
  
  flag=1;
  if (hist->records)
  { key=hist_key(hist,record,&length);
    index=hist_mask(calc_histnr(key,length ? length : hist->key_length),
                                       hist->blength,hist->records);
    do
    {
      pos= array_dynamic_element(&hist->array,index,HIST_LINK*);
      if (!histcmp(hist,pos,key,length))
      {
       	hist->current_record= index;
        pos->freq+=frequency;
	return (0);
      }
      if (flag)
      {
	flag=0;					/* Reset flag */
	if (hist_rec_mask(hist,pos,hist->blength,hist->records) != index)
	  break;				/* Wrong link */
      }
    }
    while ((index=pos->next) != NO_RECORD);
  }
  hist->current_record= NO_RECORD;
  return(hist_add_entry_frequency(hist,record,frequency));

}


         /* Ajoute 'frequency' a la frequence  lorsque  un enregistrement existe dans l'histogramme  
           ou insert un nouveau couple (record,frequency) si l'enregistrement record  n'existe pas dans l'histogramme  */
int hist_insert_frequency_index(HIST *hist,const char *record,uint frequency)
{ 
  HIST_LINK *pos,*data;
  uint hist_nr,index,length;
  char *key;
  int flag ;  
  
  flag=1;
  if (hist->records)
  { key=hist_key(hist,record,&length);
    index=hist_mask(calc_histnr(key,length ? length : hist->key_length),
                                       hist->blength,hist->records);
    do
    {
      pos= array_dynamic_element(&hist->array,index,HIST_LINK*);
      if (!histcmp_n(hist,pos,key,length+PID))
      {
       	hist->current_record= index;
        pos->freq+=frequency;
	return (0);
      }
      if (flag)
      {
	flag=0;					/* Reset flag */
	if (hist_rec_mask(hist,pos,hist->blength,hist->records) != index)
	  break;				/* Wrong link */
      }
    }
    while ((index=pos->next) != NO_RECORD);
  }
  hist->current_record= NO_RECORD;
  return(hist_add_entry_frequency(hist,record,frequency));

}










/*****************************************************************
MATTHIEU : Fonction de creation d'un nouveau bloc de type HIST_ARRAY.
Il est insere en tete de la liste chainee
*****************************************************************/
HIST_ARRAY* ha_nouveau(HIST_ARRAY* ha) {
  int soi=sizeof(int);
  HIST_ARRAY *new;

  // creation du HIST_ARRAY
  new=(HIST_ARRAY *)malloc(sizeof(HIST_ARRAY));
  // creation de data
  new->data=(char*)malloc(HIST_BUFFER_SIZE);
  // mise en place de l'offset de depart
  memcpy(new->data,&soi,soi);
  // mise en tete de la liste chainee
  new->next=ha;

  return new;
}

/*************************************************************
MATTHIEU : bete copie de  hist_array_buffer_free
*************************************************************/
void ha_libere(HIST_ARRAY** ha) {
  HIST_ARRAY *ptr;
  HIST_ARRAY *head=*ha;
  while (head!=NULL) {
    ptr=head->next;
    free(head->data);
    free(head); 
    head=ptr;       
  }
  *ha=NULL;
}

/*************************************************************
MATTHIEU : retourne 1 si ha ne contient rien, 0 sinon
*************************************************************/
int ha_estvide(HIST_ARRAY* ha) {
  int offset;
  memcpy(&offset,ha->data,sizeof(int));
  if (offset==sizeof(int)) {
    return 1;
  }else{
    return 0;
  }
}

/***************************************************************
MATTHIEU : fusion de deux HIST_ARRAY
Le premier est une queue, le second une tete
Renvoie la queue globale
***************************************************************/
HIST_ARRAY* ha_fusionne(HIST_ARRAY *ha_queue,HIST_ARRAY *ha_tete){
    HIST_ARRAY *tmp;
    if (ha_queue!=NULL) {
	tmp=ha_queue;
	tmp->next=ha_tete;
    } else {
	tmp=ha_tete;
    }
    while (tmp->next!=NULL) {
	tmp=tmp->next;
    }
    return tmp;
}

/************************************************************************
MATTHIEU : Fonction d'ajout de couple dans un hist_array binaire
************************************************************************/

HIST_ARRAY* ha_ajout_couple(HIST_ARRAY *ha,const char* str,int val){
  // raccourci vers le contenu de ha
  char* contenu;
  // longueur de la chaine
  int sl=strlen(str);
  // Recuperation de l'offset actuel dans ha
  int offset;
  memcpy(&offset,ha->data,sizeof(int));
  // on verifie si on a la place d'inserer
  if ((offset+sl+2*sizeof(int))>HIST_BUFFER_SIZE) {
    ha=ha_nouveau(ha);
    offset=sizeof(int);    
  }

  // mise en place du raccourci
  contenu=ha->data;
  // Ajout de la taille de la chaine
  memcpy(contenu+offset,&sl,sizeof(int));
  offset+=sizeof(int);
  // Ajout de la chaine
  memcpy(contenu+offset,str,sl);
  offset+=sl;
  // Ajout de la valeur associee
  memcpy(contenu+offset,&val,sizeof(int));
  offset+=sizeof(int);
  // mise a jour de l'offset
  memcpy(contenu,&offset,sizeof(int));

  return ha;
}

/************************************************************************
MATTHIEU : Fonction d'ajout de triplet dans un hist_array binaire
************************************************************************/

HIST_ARRAY* ha_ajout_triplet(HIST_ARRAY *ha,const char* str,int val1,int val2){
  // raccourci vers le contenu de ha
  char* contenu;
  // longueur de la chaine
  int sl=strlen(str);
  // Recuperation de l'offset actuel dans ha
  int offset;
  memcpy(&offset,ha->data,sizeof(int));
  //printf("ajout d'un triplet, offset = %d\n",offset);
  // on verifie si on a la place d'inserer
  if ((offset+sl+3*sizeof(int))>HIST_BUFFER_SIZE) {
    //printf("creation d'un nouveau bloc pour ajout triplet\n");
    ha=ha_nouveau(ha);
    offset=sizeof(int);    
  }

  // mise en place du raccourci
  contenu=ha->data;

  // Ajout de la taille de la chaine
  memcpy(contenu+offset,&sl,sizeof(int));
  offset+=sizeof(int);
  // Ajout de la chaine
  memcpy(contenu+offset,str,sl);
  offset+=sl;
  // Ajout de la valeur associee
  memcpy(contenu+offset,&val1,sizeof(int));
  offset+=sizeof(int);
  // Ajout de la seconde valeur associee
  memcpy(contenu+offset,&val2,sizeof(int));
  offset+=sizeof(int);
  // mise a jour de l'offset
  memcpy(contenu,&offset,sizeof(int));
  //printf("offset apres ajout : %d\n",offset);
  return ha;
}

/************************************************************************
MATTHIEU : Fonction d'extraction de couple d'un hist_array binaire
************************************************************************/

HIST_ARRAY* ha_extract_couple(HIST_ARRAY* ha,char* str,int* val,int* offset){
  // raccourci vers le contenu de ha
  char* contenu;
  // longueur de la chaine
  int sl;
  // Recuperation de l'offset actuel dans ha
  int size;
  memcpy(&size,ha->data,sizeof(int));
  // on verifie si il reste des donnees, sinon on passe au suivant
  if ((*offset)>=size) {
    ha=ha->next;
    //return NULL;
    // si on a fini, on retourne NULL
    if (ha==NULL) {
      return NULL;
    }
    (*offset)=sizeof(int);    
  }

  // mise en place du raccourci
  contenu=ha->data;
  // extraction de la taille de la chaine
  memcpy(&sl,contenu+(*offset),sizeof(int));
  (*offset)+=sizeof(int);
  // extraction de la chaine
  memcpy(str,contenu+(*offset),sl);
  str[sl]=0;
  (*offset)+=sl;
  // extraction de la valeur associee
  memcpy(val,contenu+(*offset),sizeof(int));
  (*offset)+=sizeof(int);

  return ha;
}

/************************************************************************
MATTHIEU : Fonction d'extraction de triplet d'un hist_array binaire
************************************************************************/

HIST_ARRAY* ha_extract_triplet(HIST_ARRAY *ha,char* str,int* val1,int* val2,int* offset){
  // raccourci vers le contenu de ha
  char* contenu;
  // longueur de la chaine
  int sl;
  // Recuperation de l'offset actuel dans ha
  int size;
  memcpy(&size,ha->data,sizeof(int));
  // on verifie si il reste des donnees, sinon on passe au suivant
  if ((*offset)>=size) {
    ha=ha->next;
    // si on a fini, on retourne NULL
    if (ha==NULL) {
      return NULL;
    }
    (*offset)=sizeof(int);    
    //printf("nouveau bloc\n");
  }

  // mise en place du raccourci
  contenu=ha->data;
  // extraction de la taille de la chaine
  memcpy(&sl,contenu+(*offset),sizeof(int));
  (*offset)+=sizeof(int);
  // extraction de la chaine
  memcpy(str,contenu+(*offset),sl);
  str[sl]=0;
  (*offset)+=sl;
  // extraction de la valeur associee
  memcpy(val1,contenu+(*offset),sizeof(int));
  (*offset)+=sizeof(int);
  // extraction de la seconde valeur associee
  memcpy(val2,contenu+(*offset),sizeof(int));
  (*offset)+=sizeof(int);

  return ha;
}

/************************************************************************************
MATTHIEU : transformation d'un hist en HIST_ARRAY
Renommage de hist_array_buffer
************************************************************************************/
HIST_ARRAY* HIST_to_HIST_ARRAY(HIST *hist) {
  HIST_ARRAY  *ptr=NULL;
  char        record[80],*entry;
  uint        i,records;
  HIST_LINK* data;

  if (hist!=NULL) {
    ptr=ha_nouveau(NULL);
    ptr->next=NULL;
    data=array_dynamic_element(&hist->array,0,HIST_LINK*);
    records=hist->records;
    for (i=0 ; i < records ; i++){
      ptr=ha_ajout_couple(ptr,data[i].data,data[i].freq);
    }
  }
  return ptr;
}

/************************************************************************************
MATTHIEU : transformation d'un hist en HIST_ARRAY (contenu de type triplet)
Renommage de hist_array_buffer_proc_index
************************************************************************************/
HIST_ARRAY* HIST_to_HIST_ARRAY_proc(HIST *hist,int proc) {
  HIST_ARRAY  *ptr=NULL;
  char        record[80],*entry;
  uint        i,records;
  HIST_LINK* data;
  
  if (hist!=NULL) {
    ptr=ha_nouveau(NULL);
    data=array_dynamic_element(&hist->array,0,HIST_LINK*);
    records=hist->records;
    for (i=0 ; i < records ; i++){
      ptr=ha_ajout_triplet(ptr,data[i].data,data[i].freq,proc);
    }
  }
  return ptr;
}

/***********************************************************************************************
MATTHIEU : cree un histogramme associe a une liste chainee contenant des  couples de la forme (record, frequence) 
transforme le HIST_ARRAY en HIST*
renommage de  hist_load_array_buffer
***********************************************************************************************/
void HIST_ARRAY_to_HIST(HIST_ARRAY* list, HIST* hist,uint size,uint key_offset,uint key_length,
                                      char* (*get_key)(const char *,uint*), void (*free_element)(void*)) {   
  HIST_ARRAY* ptr=list;
  int freq;
  char *buf,record[80];
  int offset=sizeof(int);
  if (list!=NULL) {
    if (hist_init(hist,size,key_offset,key_length,get_key,free_element)){
      printf("Histogramm not created ...\n");
      return;
    }   
    while ((ptr = ha_extract_couple(ptr,record,&freq,&offset)) != NULL){
	//printf(">>>>>>%s %d\n",record,freq);
      if (hist_insert_frequency(hist,record,freq)!=0){
	printf(" row not inserted ...\n");
      } 
    }
  }else{
      printf("LIST NULL\n");
  }
}

/***********************************************************************************************
MATTHIEU : cree un histogramme associe a une liste chainee contenant des  couples de la forme (record, frequence) 
transforme le HIST_ARRAY en HIST*
renommage de  hist_load_array_buffer_proc_index
***********************************************************************************************/
void HIST_ARRAY_to_HIST_proc(HIST_ARRAY* list,HIST* hist,uint size,uint key_offset,uint key_length,
                                      char* (*get_key)(const char *,uint*), void (*free_element)(void*)) {   
  uint freq,proc;
  char *buf,record[80];
  HIST_ARRAY* ptr=list;
  int offset=sizeof(int);

  if (list!=NULL) {
    if (hist_init(hist,size,key_offset,key_length,get_key,free_element)){
      printf("Histogramm not created ...\n");
      return;
    }
    while ((ptr = ha_extract_triplet(ptr,record,&freq,&proc,&offset)) != NULL){
      if (hist_insert_frequency(hist,record,freq)){
	printf(" row not inserted ...\n");
      } 
    }
  }
}



// MATTHIEU : renommage de array_buffer_length et legere modif
int ha_nb_blocs(HIST_ARRAY *list) {   
  int length=0;
   while (list!=NULL){
          list=list->next;
          length++;  
   }
   return(length);
}





/************************************* VIEILLERIES *********************************


// cree  une liste chaine  contenent  les couples (valeur, frequency) des enregistrements
//   de l'histogramme . Ces valeurs sont separees par  le separateur  '\n'                  

int hist_array_buffer(HIST *hist,HIST_ARRAY **list)
{
  HIST_LINK   *data;
  HIST_ARRAY  *ptr;
  char        buffer[HIST_BUFFER_SIZE],record[80],*entry;
  uint        i,records,length=0,reclength=0;
  
  *list=0;
  if (hist) {
      data=array_dynamic_element(&hist->array,0,HIST_LINK*);
      records=hist->records;
      for (i=0 ; i < records ; i++){

               sprintf(record,"%s %d\n",data[i].data,data[i].freq);
               reclength=strlen(record); 

	       if ((length+reclength)>= HIST_BUFFER_SIZE-1)
		 { 
                   ptr=(HIST_ARRAY*) malloc(sizeof(struct st_hist_array));
                   if (ptr==0) {
			 printf(" No enough memory ...\n");
                         return(0);
                    }
                   entry=(char*) malloc(length+1);
                   memcpy(entry,buffer,length); entry[length]=0;
                   ptr->data=entry;
                   ptr->next=*list;
                   *list=ptr;
                   length=0;
                 }            
                memcpy(buffer+length,record,reclength); length+=reclength;                
	       
    }
    if (length!=0) {
                 ptr=(HIST_ARRAY*) malloc(sizeof(struct st_hist_array));
                 if (ptr==0) {
			 printf(" No enough memory ...\n");
                         return(0);
                 }
                 entry=(char*) malloc(length+1);
                 memcpy(entry,buffer,length); entry[length]=0;
                 ptr->data=entry;
                 ptr->next=*list;
                 *list=ptr;
                 length=0; reclength=0;
    }
  }
  return(1);
}

// cree  une liste chaine  contenent  des triplets (valeur, frequency,proc) des enregistrements
//   de l'histogramme . Ces valeurs sont separees par  le  caractere  '\n'                  
int hist_array_buffer_proc_index(HIST *hist,HIST_ARRAY **list,int proc)
{
  HIST_LINK   *data;
  HIST_ARRAY  *ptr;
  char        buffer[HIST_BUFFER_SIZE],record[80],*entry;
  uint        i,records,length=0,reclength=0;
  
  *list=0;
  if (hist) {
      data=array_dynamic_element(&hist->array,0,HIST_LINK*);
      records=hist->records;
      for (i=0 ; i < records ; i++) {
               sprintf(record,"%s %d %d\n",data[i].data,data[i].freq,proc);
               reclength=strlen(record); 

	       if ((length+reclength)>= HIST_BUFFER_SIZE-1) { 
                   ptr=(HIST_ARRAY*) malloc(sizeof(struct st_hist_array));
                   if (ptr==0)
                       {
			 printf(" No enough memory ...\n");
                         return(0);
                       }
                   entry=(char*) malloc(length+1);
                   memcpy(entry,buffer,length); entry[length]=0;
                   ptr->data=entry;
                   ptr->next=*list;
                   *list=ptr;
                   length=0;
                 }
               memcpy(buffer+length,record,reclength); length+=reclength;
    }
    if (length!=0) {
                 ptr=(HIST_ARRAY*) malloc(sizeof(struct st_hist_array));
                 if (ptr==0)  {
			 printf(" No enough memory ...\n");
                         return(0);
                 }
                 entry=(char*) malloc(length+1);
                 memcpy(entry,buffer,length); entry[length]=0;
                 ptr->data=entry;
                 ptr->next=*list;
                 *list=ptr;
                 length=0; reclength=0;
    }
  }
  return(1);
}


// renvoi le nombre d'elements dans d'une liste chainee 
int array_buffer_length(HIST_ARRAY *list)
{   
  int length=0;

   while (list){
          list=list->next;
          length++;  
   }
   return(length);
}

// copie une liste chainee dans une autre liste chainee 
int array_buffer_copy(HIST_ARRAY *list1,HIST_ARRAY** list2)
{ 
  char *data;  
  int length=0;
  HIST_ARRAY *ptr;
  
   *list2=0;
   while (list1){
          ptr=(HIST_ARRAY*) malloc(sizeof(struct st_hist_array));
           if (ptr==0)  {
			 printf(" No enough memory ...\n");
                         return(0);
                 }
           length=strlen(list1->data)+1;
           data=(char*) malloc(length*sizeof(char));
           memcpy(data,list1->data,length);
           ptr->data=data;
   	   ptr->next=*list2;
           *list2=ptr; 
           list1=list1->next;
         
   }
   return(1);
}

// cree un histogramme associe a une liste chainee contenant des  couples de la forme (record, frequence) 
//transforme le HIST_ARRAY en HIST* 
int hist_load_array_buffer(HIST_ARRAY *list,HIST *hist,uint size,uint key_offset,uint key_length,
                                      char* (*get_key)(const char *,uint*), void (*free_element)(void*))
{   
  uint freq;
  char *buf,record[80];


    if (hist_init(hist,size,key_offset,key_length,get_key,free_element)){
        printf("Histogramm not created ...\n");
        return(0);
    }
    while (list){
          buf=list->data;
          while (strlen(buf)) {
                sscanf(buf,"%s %d",record,&freq); buf=(char*) strchr(buf,SEP_CHAR)+SEP_LENGTH;
                if (hist_insert_frequency(hist,record,freq)){
                    printf(" row not inserted ...\n");
                } 
          }
          list=list->next;  
   }
   return(1);
}


// cree un histogramme associe a une liste chainee contenant des triplets de la forme (record,frequence,indice) 
int hist_load_array_buffer_proc_index(HIST_ARRAY *list,HIST *hist,uint size,uint key_offset,uint key_length,
                                      char* (*get_key)(const char *,uint*), void (*free_element)(void*))
{   
  uint freq,proc;
  char *buf,record[80], entry[80];

    if (hist_init(hist,size,key_offset,key_length,get_key,free_element)){
        printf("Histogramm not created ...\n");
        return(0);
    }
    while (list){
          buf=list->data;
          while (strlen(buf)) {
                sscanf(buf,"%s %d %d",record,&freq, &proc); buf=(char*) strchr(buf,SEP_CHAR)+SEP_LENGTH;
                if (hist_insert_frequency(hist,record,freq)){
                    printf(" row not inserted ...\n");
                } 
          }
          list=list->next;  
   }
   return(1);
}
************************************************************ fin vieilleries****************************/

//  Ajoute a un histogramme une liste de couples de la forme (record, frequence) stockee dans dans buffer 
int hist_append_buffer(char* buffer,HIST *hist)
{   
  uint freq;
  char *buf,record[80];

          buf=buffer;
          while (strlen(buf)) {
                sscanf(buf,"%s %d",record,&freq); buf=(char*) strchr(buf,SEP_CHAR)+SEP_LENGTH;
                if (hist_insert_frequency(hist,record,freq)) {
                    printf(" row not inserted ...\n");
                    return(0);
                } 
          }
          return(1);
}

int old_hist_append_buffer_proc_index(char* buffer,HIST *hist)
{   
  uint freq,proc;
  char *buf,record[80],entry[80];

          buf=buffer;
          while (strlen(buf)) {
                sscanf(buf,"%s %d %d",record,&freq,&proc); buf=(char*) strchr(buf,SEP_CHAR)+SEP_LENGTH;
                sprintf(entry,"%s:%4d",record,proc);
                if (hist_insert_frequency(hist,entry,freq)) {
                    printf(" row not inserted ...\n");
                    return(0);
                } 
          }
          return(1);
}

void ha_show (HIST_ARRAY* list) {   
  HIST_ARRAY* ptr=list;
  int freq;
  char record[80];
  int offset=sizeof(int);
  if (list==NULL) {
      printf("histogramme vide!!!!!\n");
      return;
  }   	
  while ((ptr = ha_extract_couple(ptr,record,&freq,&offset)) != NULL){
	printf("show : %s %d\n",record,freq);  
  } 	
}

void ha_show_proc (HIST_ARRAY* list) {   
  HIST_ARRAY* ptr=list;
  int freq,proc;
  char record[80];
  int offset=sizeof(int);
  if (list==NULL) {
      printf("histogramme vide!!!!!\n");
      return;
  }   	
  while ((ptr =ha_extract_triplet(ptr,record,&freq,&proc,&offset)) != NULL){
	printf("showp : %s %d %d\n",record,freq,proc);  
  } 	
}



/****************************
int hist_array_buffer_show(HIST_ARRAY *list)
{
  int i=0;
    while (list) { 
      printf("####### Bloc  number : %d ,size : %d \n %s \n",i,strlen(list->data),list->data);  
      list=list->next;  i++;      
    }
  //scanf("\n%d",&i);
}


int hist_array_buffer_show_file(HIST_ARRAY *list,char* file)
{
  int i=0,j;
  FILE *fd;
 
  fd=fopen(file,"w");
  while (list)
    { i++;
      fprintf(fd,"####### Bloc  number : %d ,size : %d \n %s \n",i,strlen(list->data),list->data);  
      list=list->next;       
    }
  fclose(fd);
}

int hist_array_buffer_free(HIST_ARRAY **list) { 
 HIST_ARRAY *ptr;

  while (*list)
    { ptr=(*list)->next;
      free((*list)->data);
      free(*list); 
      *list=ptr;       
    }
  *list=0;
}
**********/


int hist_copy(HIST *hist1,HIST *hist2)
{
char *record,*recpos;
uint   write_count,reclength;
ulong  pos;

     if (hist_init(hist2,hist1->records,hist1->key_offset,
               hist1->key_length, hist1->get_key, hist1->free)) {
       printf("\n histogram not copied !!!!. ");
       return(0);
     }
     pos=0;
     while (recpos=hist_element(hist1,pos))
     {  
        
        if (hist_add_entry_frequency(hist2,recpos,array_dynamic_element(&hist1->array,pos,HIST_LINK*)->freq))
           {
             printf("Got error when inserting record: %s",recpos);
            /* goto err;*/
           }
        pos++;
     }
     return(1);
}


int hist_merge(HIST *hist1,HIST *hist2)
{
char *record,*recpos;
uint   write_count,reclength=120;
ulong  pos;

     pos=0;
     while (recpos=hist_element(hist1,pos))
     {  
         if (hist_add_entry_frequency(hist2,recpos,array_dynamic_element(&hist1->array,pos,HIST_LINK*)->freq))
           {
             printf("Got error when inserting record : %s",recpos);
             return(0);
           } 
       pos++;
       }
     return(1);
}


int hist_comp(HIST *hist1,HIST *hist2)
{

  uint i,records,frequency,length;
  HIST_LINK *data=array_dynamic_element(&hist1->array,0,HIST_LINK*);
  char *key;

    
    if (hist1->records != hist2->records) return(0);
    for (i=0,records=hist1->records ; i < records ; i++){
      //key=hist_key(hist2,data[i].data,&length);
      length=strlen(data[i].data);
      if (hist_search_frequency(hist2,data[i].data,length,&frequency)==0){
	     printf(" record : %s  not found ...\n", data[i].data);
	     return(0);
      } 
          else
	    if (frequency !=data[i].freq) return(0);
    }
    return(1);

}

int hist_comp_n(HIST *hist1,HIST *hist2)
{

  uint i,records,frequency,length;
  HIST_LINK *data=array_dynamic_element(&hist1->array,0,HIST_LINK*);
  char *key;

    
    if (hist1->records != hist2->records) return(0);
    for (i=0,records=hist1->records ; i < records ; i++){
      key=hist_key(hist1,data[i].data,&length);
      //length=strlen(data[i].data);
      if (hist_search_frequency_n(hist2,key,length,&frequency)==0){
	     printf(" record : %s  not found ...\n", data[i].data);
	     return(0);
      } 
          else
	    if (frequency !=data[i].freq) return(0);
    }
    return(1);

}

int hist_comp_n2(HIST *hist1,HIST *hist2)
{

  uint i,records,frequency,length;
  HIST_LINK *data=array_dynamic_element(&hist1->array,0,HIST_LINK*);
  char *key;
  int erreur=1;
    
  // if (hist1->records != hist2->records) return(0);
    for (i=0,records=hist1->records ; i < records ; i++){
	//key=hist_key(hist1,data[i].data,&length);
      length=strlen(data[i].data);
      if (hist_search_frequency_n(hist2,data[i].data,length,&frequency)==0){
	  //printf(" ** record : '%s'   not found ...\n", data[i].data);
	     erreur=0;
      } 
          else
	    if (frequency !=data[i].freq){
		//	printf(" ** record : '%s' : freq 1 =  %d  ///  freq 2 = %d ...\n", data[i].data,data[i].freq,frequency);
		//return(0);
		erreur=0;
	    }
    }
    return erreur;

}






int hist_select(HIST *hist1,uint frequency,HIST *hist2)
{
char *record,*recpos;
uint   write_count,reclength;
ulong  pos;

     if (hist_init(hist2,0,hist1->key_offset,
               hist1->key_length, hist1->get_key, hist1->free)) {
       printf("\n histogram not created !!!!. ");
       return(0);
     }
     pos=0;
     while (recpos=hist_element(hist1,pos))
     {  
        if (array_dynamic_element(&hist1->array,pos,HIST_LINK*)->freq>=frequency) {
        
            if (hist_add_entry_frequency(hist2,recpos,array_dynamic_element(&hist1->array,pos,HIST_LINK*)->freq))
               {
                  printf("Got error when inserting record : %s",record);
                  return(0);
                  /* goto err;*/
            }
	  }
        pos++;
     }
     return(1);
}


int hist_intersect(HIST *hist1,HIST *hist2,HIST *hist3)
{
char *record,*recpos;
uint   write_count,reclength;
ulong  pos;

   if (hist_init(hist3,0,hist1->key_offset,
               hist1->key_length, hist1->get_key, hist1->free)) {
       printf("\n histogram not copied !!!!. ");
       return(0);
     }
     pos=0;
     while (recpos=hist_element(hist1,pos))
     {  
        if (hist_search(hist2,recpos,hist2->key_length)){
             if (hist_insert_frequency(hist3,recpos,0)) {
                printf("Got error when inserting record: %s",recpos);
                return(0);
	      }
	  }
       pos++;
       }
     return(1);
}



/***************************************************************************
Fournit l'intersection des deux histogrammes
PARAMETRES :
hist1 (E) : l'histogramme de la relation 1
hist2 (E) : l'histogramme de la relation 2
hist3 (S) : un histogramme correspondant aux valeurs prsentent dans les deux relations, precisant en outre la configuration vis a vis du seuil
threshold (E) : le seuil
***************************************************************************/
int hist_intersect_index(HIST *hist1,HIST *hist2,HIST *hist3,int threshold)
{
uint       size=0,length,freq1,freq2,i,ind;
HIST_LINK *data;

     if ((hist1->records)>(hist2->records)){
         size=hist2->records;
         data=array_dynamic_element(&hist2->array,0,HIST_LINK*);
         if  (hist_init(hist3,size,hist1->key_offset,
              hist1->key_length, hist1->get_key,hist1->free)) {
              printf("\n histogram not copied !!!!. ");
              return(1);
	    }
         for (i=0;i<size;i++){
              length=strlen(data[i].data); 
               
              if (hist_search_frequency(hist1,data[i].data,length,&freq1)){
                  freq2=data[i].freq;                 
                  if ((freq1<threshold) && (freq2<threshold)) ind=HASH_INDEX;
                  else ind=(freq1<freq2)?LEFT_INDEX:RIGHT_INDEX;
 
                  if (hist_add_entry_frequency(hist3,data[i].data,ind)) {
                     printf("Got error when inserting record: %*s",data[i].data);
 		   }
                   
		}
	    }
       }
      else {
         size=hist1->records;
         data=array_dynamic_element(&hist1->array,0,HIST_LINK*);
         if  (hist_init(hist3,size,hist1->key_offset,
              hist1->key_length, hist1->get_key,hist1->free)) {
              printf("\n histogram not copied !!!!. ");
              return(1);
	    }
         for (i=0;i<size;i++){
              length=strlen(data[i].data); 
              if (hist_search_frequency(hist2,data[i].data,length,&freq2)){
                  freq1=data[i].freq;
                 
                  if ((freq1<threshold) && (freq2<threshold)) ind=HASH_INDEX;
                  else ind=(freq1<freq2)?LEFT_INDEX:RIGHT_INDEX;
           
                  if (hist_add_entry_frequency(hist3,data[i].data,ind)) {
                     printf("Got error when inserting record: %*s",data[i].data);
                     /* goto err;*/
		   } 
		}
	    }
       }
     return(0);
}


int hist_intersect_index_double_frequencies(HIST *hist1,HIST *hist2,HIST *hist3,int threshold)
{
uint       size=0,length,freq1,freq2,i,ind;
char       entry[512];
HIST_LINK *data;


     
   
     if ((hist1->records)>(hist2->records)){
         size=hist2->records;
         data=array_dynamic_element(&hist2->array,0,HIST_LINK*);
         if  (hist_init(hist3,size,hist1->key_offset,
              hist1->key_length, get_trunc_key_double_int,hist1->free)) {
              printf("\n histogram not copied !!!!. ");
              return(1);
	    }
         for (i=0;i<size;i++){
              length=strlen(data[i].data); 
              
              if (hist_search_frequency(hist1,data[i].data,length,&freq1)){
                  freq2=data[i].freq;                 
                  if ((freq1<threshold) && (freq2<threshold)) ind=HASH_INDEX;
                  else ind=(freq1<freq2)?LEFT_INDEX:RIGHT_INDEX;
                  sprintf(entry,"%s %*d %*d",data[i].data,FREQ_SPACE,freq1,FREQ_SPACE,freq2);
                  if (hist_add_entry_frequency(hist3,entry,ind)) {
                     printf("Got error when inserting record: %*s",data[i].data);
 		   }
                   
		}
	    }
       }
      else {
         size=hist1->records;
         data=array_dynamic_element(&hist1->array,0,HIST_LINK*);
         if  (hist_init(hist3,size,hist1->key_offset,
              hist1->key_length,get_trunc_key_double_int,hist1->free)) {
              printf("\n histogram not copied !!!!. ");
              return(1);
	    }
         for (i=0;i<size;i++){
              length=strlen(data[i].data); 
              if (hist_search_frequency(hist2,data[i].data,length,&freq2)){
                  freq1=data[i].freq;
                 
                  if ((freq1<threshold) && (freq2<threshold)) ind=HASH_INDEX;
                  else ind=(freq1<freq2)?LEFT_INDEX:RIGHT_INDEX;
                  sprintf(entry,"%s %*d %*d",data[i].data,FREQ_SPACE,freq1,FREQ_SPACE,freq2);
                  if (hist_add_entry_frequency(hist3,entry,ind)) {
                     printf("Got error when inserting record: %*s",data[i].data);
                     /* goto err;*/
		   } 
		}
	    }
       }
     return(0);
}

/* Amlioration des temps d'exec par Matthieu */
int hist_intersect_index_freq1(HIST *hist1,HIST *hist2,HIST *hist3,int threshold)
{
uint       size=0,length,freq1,freq2,i,ind;
char       entry[512];
HIST_LINK *data;

     if ((hist1->records)>(hist2->records)){
         size=hist2->records;
         data=array_dynamic_element(&hist2->array,0,HIST_LINK*);
         if  (hist_init(hist3,size,hist1->key_offset,
			//              hist1->key_length, get_trunc_key_double_int,hist1->free)) {
			hist1->key_length, get_trunc_key_simple_int,hist1->free)) {
              printf("\n histogram not copied !!!!. ");
              return(1);
	    }
         for (i=0;i<size;i++){
              length=strlen(data[i].data); 
              
              if (hist_search_frequency(hist1,data[i].data,length,&freq1)){
                  freq2=data[i].freq;                 
                  if ((freq1<threshold) && (freq2<threshold)) ind=HASH_INDEX;
                  else ind=(freq1<freq2)?LEFT_INDEX:RIGHT_INDEX;
                  sprintf(entry,"%s %*d",data[i].data,FREQ_SPACE,freq1);
                  if (hist_add_entry_frequency(hist3,entry,ind)) {
                     printf("Got error when inserting record: %*s",data[i].data);
 		   }
                   
		}
	    }
       }
      else {
         size=hist1->records;
         data=array_dynamic_element(&hist1->array,0,HIST_LINK*);
         if  (hist_init(hist3,size,hist1->key_offset,
			// hist1->key_length,get_trunc_key_double_int,hist1->free)) {
              hist1->key_length,get_trunc_key_simple_int,hist1->free)) {
              printf("\n histogram not copied !!!!. ");
              return(1);
	    }
         for (i=0;i<size;i++){
              length=strlen(data[i].data); 
              if (hist_search_frequency(hist2,data[i].data,length,&freq2)){
                  freq1=data[i].freq;
                 
                  if ((freq1<threshold) && (freq2<threshold)) ind=HASH_INDEX;
                  else ind=(freq1<freq2)?LEFT_INDEX:RIGHT_INDEX;
                  sprintf(entry,"%s %*d",data[i].data,FREQ_SPACE,freq1);
                  if (hist_add_entry_frequency(hist3,entry,ind)) {
                     printf("Got error when inserting record: %*s",data[i].data);
                     /* goto err;*/
		   } 
		}
	    }
       }
     return(0);
}

/* Amlioration des temps d'exec par Matthieu */
int hist_intersect_index_freq2(HIST *hist1,HIST *hist2,HIST *hist3,int threshold)
{
uint       size=0,length,freq1,freq2,i,ind;
char       entry[512];
HIST_LINK *data;


     
   
     if ((hist1->records)>(hist2->records)){
         size=hist2->records;
         data=array_dynamic_element(&hist2->array,0,HIST_LINK*);
         if  (hist_init(hist3,size,hist1->key_offset,
              hist1->key_length, get_trunc_key_simple_int,hist1->free)) {
              printf("\n histogram not copied !!!!. ");
              return(1);
	    }
         for (i=0;i<size;i++){
              length=strlen(data[i].data); 
              
              if (hist_search_frequency(hist1,data[i].data,length,&freq1)){
                  freq2=data[i].freq;                 
                  if ((freq1<threshold) && (freq2<threshold)) ind=HASH_INDEX;
                  else ind=(freq1<freq2)?LEFT_INDEX:RIGHT_INDEX;
                  sprintf(entry,"%s %*d",data[i].data,FREQ_SPACE,freq2);
                  if (hist_add_entry_frequency(hist3,entry,ind)) {
                     printf("Got error when inserting record: %*s",data[i].data);
 		   }
                   
		}
	    }
       }
      else {
         size=hist1->records;
         data=array_dynamic_element(&hist1->array,0,HIST_LINK*);
         if  (hist_init(hist3,size,hist1->key_offset,
			//hist1->key_length,get_trunc_key_double_int,hist1->free)) {
              hist1->key_length,get_trunc_key_simple_int,hist1->free)) {
              printf("\n histogram not copied !!!!. ");
              return(1);
	    }
         for (i=0;i<size;i++){
              length=strlen(data[i].data); 
              if (hist_search_frequency(hist2,data[i].data,length,&freq2)){
                  freq1=data[i].freq;
                 
                  if ((freq1<threshold) && (freq2<threshold)) ind=HASH_INDEX;
                  else ind=(freq1<freq2)?LEFT_INDEX:RIGHT_INDEX;
                  sprintf(entry,"%s %*d",data[i].data,FREQ_SPACE,freq2);
                  if (hist_add_entry_frequency(hist3,entry,ind)) {
                     printf("Got error when inserting record: %*s",data[i].data);
                     /* goto err;*/
		   } 
		}
	    }
       }
     return(0);
}

/******************************************************************************
** Remove one record from hist-table. The record with the same record
** ptr is removed.
** if there is a free-function it's called for record if found
******************************************************************************/

int hist_delete(HIST *hist,char *record)
{
  uint blength,pos2,pos_histnr,lastpos_histnr,index,empty_index;
  HIST_LINK *data,*lastpos,*gpos,*pos,*pos3,*empty;
 
  blength=hist->blength;
  data=array_dynamic_element(&hist->array,0,HIST_LINK*);

  /* Search after record with key */
  pos=data+ hist_mask(rec_histnr(hist,record),blength,hist->records);
  gpos = 0;

  while (pos->data != record)
  {
    gpos=pos;
    if (pos->next == NO_RECORD)
      return(1);			/* Key not found */
    pos=data+pos->next;
  }

  if ( --(hist->records) < hist->blength >> 1) hist->blength>>=1;
  hist->current_record= NO_RECORD;
  lastpos=data+hist->records;

  /* Remove link to record */
  empty=pos; empty_index=(uint) (empty-data);
  if (gpos)
    gpos->next=pos->next;		/* unlink current ptr */
  else if (pos->next != NO_RECORD)
  {
    empty=data+(empty_index=pos->next);
    pos->data=empty->data;
    pos->next=empty->next;
  }

  if (empty == lastpos)			/* last key at wrong pos or no next link */
    goto exit;

  /* Move the last key (lastpos) */
  lastpos_histnr=rec_histnr(hist,lastpos->data);
  /* pos is where lastpos should be */
  pos=data+hist_mask(lastpos_histnr,hist->blength,hist->records);
  if (pos == empty)			/* Move to empty position. */
  {
    empty[0]=lastpos[0];
    goto exit;
  }
  pos_histnr=rec_histnr(hist,pos->data);
  /* pos3 is where the pos should be */
  pos3= data+hist_mask(pos_histnr,hist->blength,hist->records);
  if (pos != pos3)
  {					/* pos is on wrong posit */
    empty[0]=pos[0];			/* Save it here */
    pos[0]=lastpos[0];			/* This should be here */
    hist_movelink(data,(uint) (pos-data),(uint) (pos3-data),empty_index);
    goto exit;
  }
  pos2= hist_mask(lastpos_histnr,blength,hist->records+1);
  if (pos2 == hist_mask(pos_histnr,blength,hist->records+1))
  {					/* Identical key-positions */
    if (pos2 != hist->records)
    {
      empty[0]=lastpos[0];
      hist_movelink(data,(uint) (lastpos-data),(uint) (pos-data),empty_index);
      goto exit;
    }
    index= (uint) (pos-data);		/* Link pos->next after lastpos */
  }
  else index= NO_RECORD;		/* Different positions merge */

  empty[0]=lastpos[0];
  hist_movelink(data,index,empty_index,pos->next);
  pos->next=empty_index;

exit:
  pop_dynamic(&hist->array);
  if (hist->free)
    (*hist->free)((char*) record);
  return(0);
}

	/*
	  Update keys when record has changed.
	  This is much more efficent than using a delete & insert.
	  */

int hist_update(HIST *hist,char *record,char *old_key,uint old_key_length)
{
  uint index,new_index,new_pos_index,blength,records,empty;
  HIST_LINK org_link,*data,*previous,*pos;
  

  data=array_dynamic_element(&hist->array,0,HIST_LINK*);
  blength=hist->blength; records=hist->records;

  /* Search after record with key */

  index=hist_mask(calc_histnr(old_key,(old_key_length ? old_key_length :
				       hist->key_length)),
		  blength,records);
  new_index=hist_mask(rec_histnr(hist,record),blength,records);
  if (index == new_index)
    return(0);			/* Nothing to do (No record check) */
  previous=0;
  for (;;)
  {

    if ((pos= data+index)->data == record)
      break;
    previous=pos;
    if ((index=pos->next) == NO_RECORD)
      return(1);			/* Not found in links */
  }
  hist->current_record= NO_RECORD;
  org_link= *pos;
  empty=index;

  /* Relink record from current chain */

  if (!previous)
  {
    if (pos->next != NO_RECORD)
    {
      empty=pos->next;
      *pos= data[pos->next];
    }
  }
  else
    previous->next=pos->next;		/* unlink pos */

  /* Move data to correct position */
  pos=data+new_index;
  new_pos_index=hist_rec_mask(hist,pos,blength,records);
  if (new_index != new_pos_index)
  {					/* Other record in wrong position */
    data[empty] = *pos;
    hist_movelink(data,new_index,new_pos_index,empty);
    org_link.next=NO_RECORD;
    data[new_index]= org_link;
  }
  else
  {					/* Link in chain at right position */
    org_link.next=data[new_index].next;
    data[empty]=org_link;
    data[new_index].next=empty;
  }
  return(0);
}



void hist_show(HIST *hist)
{
  uint i,records,sum=0;
  HIST_LINK *data=array_dynamic_element(&hist->array,0,HIST_LINK*);

    for (i=0,records=hist->records ; i < records ; i++){
      //printf("\n row : %d -- %s -- freq : %d",i,data[i].data,data[i].freq);
      printf("%s %d\n",data[i].data,data[i].freq);
      sum+=data[i].freq;
    }
    printf("\n histogram records  : %d", hist->records);
    printf("\n frequencies sum    : %d\n", sum);
    // scanf("\n%d",&i);
}

void hist_show_file(HIST *hist,char* file)
{
  uint i,records,sum=0;
  HIST_LINK *data=array_dynamic_element(&hist->array,0,HIST_LINK*);
  FILE *fd;


     fd=fopen(file,"w");
    for (i=0,records=hist->records ; i < records ; i++){
      //printf("\n row : %d -- %s -- freq : %d",i,data[i].data,data[i].freq);
      fprintf(fd,"%s %d\n",data[i].data,data[i].freq);
      sum+=data[i].freq;
    }
    fprintf(fd,"\n histogram records  : %d", hist->records);
    fprintf(fd,"\n frequencies sum    : %d\n", sum);
    fclose(fd);
    // scanf("\n%d",&i);
}


/* Ajoute a l'histogramme un couple de la forme (record,1)  meme si un  autre entregitrement 
   existe deja l'histogramme  ayant la meme cle que record                         */
int hist_add_entry(HIST *info,const char *record)
{
  int flag;
  uint halfbuff,hist_nr,first_index,index,
       freq,freq2;
  int reclength;
  char *ptr_to_rec,*ptr_to_rec2,*entry;
  HIST_LINK *data,*empty,*gpos,*gpos2,*pos;

  init_ptr(gpos); init_ptr(gpos2);
  init_ptr(ptr_to_rec); init_ptr(ptr_to_rec2);
  

  reclength=strlen(record)+1;
  entry=(char*) malloc(reclength);
  memcpy(entry,record,reclength);
  //strcpy(entry,record);
  flag=0;
  if (!(empty=(HIST_LINK*) alloc_dynamic(&info->array)))
    return(1);				/* No more memory */

  info->current_record= NO_RECORD;
  data=array_dynamic_element(&info->array,0,HIST_LINK*);
  halfbuff= info->blength >> 1;

  index=first_index=info->records-halfbuff;
  if (index != info->records)				/* If some records */
  {
    do
    {
      pos=data+index;
      hist_nr=rec_histnr(info,pos->data);
      if (flag == 0)				/* First loop; Check if ok */
	if (hist_mask(hist_nr,info->blength,info->records) != first_index)
	  break;
      if (!(hist_nr & halfbuff))
      {						/* Key will not move */
	if (!(flag & LOWFIND))
	{
	  if (flag & HIGHFIND)
	  {
	    flag=LOWFIND | HIGHFIND;
	    /* key shall be moved to the current empty position */
	    gpos=empty;
	    ptr_to_rec=pos->data;
            freq=pos->freq;
	    empty=pos;				/* This place is now free */
	  }
	  else
	  {
	    flag=LOWFIND | LOWUSED;		/* key isn't changed */
	    gpos=pos;
	    ptr_to_rec=pos->data;
            freq=pos->freq;
	  }
	}
	else
	{
	  if (!(flag & LOWUSED))
	  {
	    /* Change link of previous LOW-key */
	    gpos->data=ptr_to_rec;
	    gpos->next=(uint) (pos-data);
            gpos->freq=freq;
	    flag= (flag & HIGHFIND) | (LOWFIND | LOWUSED);
	  }
	  gpos=pos;
	  ptr_to_rec=pos->data;
          freq=pos->freq;
	}
      }
      else
      {						/* key will be moved */
	if (!(flag & HIGHFIND))
	{
	  flag= (flag & LOWFIND) | HIGHFIND;
	  /* key shall be moved to the last (empty) position */
	  gpos2 = empty; empty=pos;
	  ptr_to_rec2=pos->data;
          freq2=pos->freq;
	}
	else
	{
	  if (!(flag & HIGHUSED))
	  {
	    /* Change link of previous hist-key and save */
	    gpos2->data=ptr_to_rec2;
	    gpos2->next=(uint) (pos-data);
            gpos2->freq=freq2;
	    flag= (flag & LOWFIND) | (HIGHFIND | HIGHUSED);
	  }
	  gpos2=pos;
	  ptr_to_rec2=pos->data;
          freq2=pos->freq;
	}
      }
    }
    while ((index=pos->next) != NO_RECORD);

    if ((flag & (LOWFIND | LOWUSED)) == LOWFIND)
    {
      gpos->data=ptr_to_rec;
      gpos->next=NO_RECORD;
      gpos->freq=freq;
    }
    if ((flag & (HIGHFIND | HIGHUSED)) == HIGHFIND)
    {
      gpos2->data=ptr_to_rec2;
      gpos2->next=NO_RECORD;
      gpos2->freq=freq2;
    }
  }
  /* Check if we are at the empty position */

  index=hist_mask(rec_histnr(info,entry),info->blength,info->records+1);
  pos=data+index;
  if (pos == empty)
  {
    pos->data=(char*) entry;
    pos->next=NO_RECORD;
    pos->freq=1;
  }
  else
  {
    /* Check if more records in same hist-nr family */
    empty[0]=pos[0];
    gpos=data+hist_rec_mask(info,pos,info->blength,info->records+1);
    if (pos == gpos)
    {
      pos->data=(char*) entry; 
      pos->next=(uint) (empty - data);
      pos->freq=1;
    }
    else
    {
      pos->data=(char*) entry;
      pos->next=NO_RECORD;
      pos->freq=1;
      hist_movelink(data,(uint) (pos-data),(uint) (gpos-data),(uint) (empty-data));
    }
  }
  if (++info->records == info->blength)
    info->blength+= info->blength;
  return(0);
}




/* Ajoute a l'histogramme un couple de la forme (record,frequency) meme si un autre entregitrement 
   existe deja l'histogramme  ayant la meme cle que record                         */
int hist_add_entry_frequency(HIST *info,const char *record,uint frequency)
{
  int flag;
  uint halfbuff,hist_nr,first_index,index,
       freq,freq2,reclength;
  char *ptr_to_rec,*ptr_to_rec2,*entry;
  HIST_LINK *data,*empty,*gpos,*gpos2,*pos;

  init_ptr(gpos); init_ptr(gpos2);
  init_ptr(ptr_to_rec); init_ptr(ptr_to_rec2);
  
  reclength=strlen(record)+1;
  entry=(char*) malloc(reclength);
  memcpy(entry,record,reclength);
  //strcpy(entry,record);
 
  flag=0;
  if (!(empty=(HIST_LINK*) alloc_dynamic(&info->array)))
    return(1);				/* No more memory */

  info->current_record= NO_RECORD;
  data=array_dynamic_element(&info->array,0,HIST_LINK*);
  halfbuff= info->blength >> 1;

  index=first_index=info->records-halfbuff;
  if (index != info->records)				/* If some records */
  {
    do
    {
      pos=data+index;
      hist_nr=rec_histnr(info,pos->data);
      if (flag == 0)				/* First loop; Check if ok */
	if (hist_mask(hist_nr,info->blength,info->records) != first_index)
	  break;
      if (!(hist_nr & halfbuff))
      {						/* Key will not move */
	if (!(flag & LOWFIND))
	{
	  if (flag & HIGHFIND)
	  {
	    flag=LOWFIND | HIGHFIND;
	    /* key shall be moved to the current empty position */
	    gpos=empty;
	    ptr_to_rec=pos->data;
            freq=pos->freq;
	    empty=pos;				/* This place is now free */
	  }
	  else
	  {
	    flag=LOWFIND | LOWUSED;		/* key isn't changed */
	    gpos=pos;
	    ptr_to_rec=pos->data;
            freq=pos->freq;
	  }
	}
	else
	{
	  if (!(flag & LOWUSED))
	  {
	    /* Change link of previous LOW-key */
	    gpos->data=ptr_to_rec;
	    gpos->next=(uint) (pos-data);
            gpos->freq=freq;
	    flag= (flag & HIGHFIND) | (LOWFIND | LOWUSED);
	  }
	  gpos=pos;
	  ptr_to_rec=pos->data;
          freq=pos->freq;
	}
      }
      else
      {						/* key will be moved */
	if (!(flag & HIGHFIND))
	{
	  flag= (flag & LOWFIND) | HIGHFIND;
	  /* key shall be moved to the last (empty) position */
	  gpos2 = empty; empty=pos;
	  ptr_to_rec2=pos->data;
          freq2=pos->freq;
	}
	else
	{
	  if (!(flag & HIGHUSED))
	  {
	    /* Change link of previous hist-key and save */
	    gpos2->data=ptr_to_rec2;
	    gpos2->next=(uint) (pos-data);
            gpos2->freq=freq2;
	    flag= (flag & LOWFIND) | (HIGHFIND | HIGHUSED);
	  }
	  gpos2=pos;
	  ptr_to_rec2=pos->data;
          freq2=pos->freq;
	}
      }
    }
    while ((index=pos->next) != NO_RECORD);

    if ((flag & (LOWFIND | LOWUSED)) == LOWFIND)
    {
      gpos->data=ptr_to_rec;
      gpos->next=NO_RECORD;
      gpos->freq=freq;
    }
    if ((flag & (HIGHFIND | HIGHUSED)) == HIGHFIND)
    {
      gpos2->data=ptr_to_rec2;
      gpos2->next=NO_RECORD;
      gpos2->freq=freq2;
    }
  }
  /* Check if we are at the empty position */

  index=hist_mask(rec_histnr(info,entry),info->blength,info->records+1);
  pos=data+index;
  if (pos == empty)
  {
    pos->data=(char*) entry;
    pos->next=NO_RECORD;
    pos->freq=frequency;
  }
  else
  {
    /* Check if more records in same hist-nr family */
    empty[0]=pos[0];
    gpos=data+hist_rec_mask(info,pos,info->blength,info->records+1);
    if (pos == gpos)
    {
      pos->data=(char*) entry;
      pos->next=(uint) (empty - data);
      pos->freq=frequency;
    }
    else
    {
      pos->data=(char*) entry;
      pos->next=NO_RECORD;
      pos->freq=frequency;
      hist_movelink(data,(uint) (pos-data),(uint) (gpos-data),(uint) (empty-data));
    }
  }
  if (++info->records == info->blength)
    info->blength+= info->blength;
  return(0);
}


/*
typedef struct hist_dynamic_array {
  char *buffer;
  uint elements,max_element;
  uint alloc_increment;
  uint size_of_element;
} HIST_DYNAMIC_ARRAY;


typedef struct st_hist {
  uint key_offset,key_length;		  // Length of key if const length 
  uint records,blength,current_record;
  HIST_DYNAMIC_ARRAY array;		  // Place for hist_keys
  char *(*get_key)(const char *,uint*);
  void (*free)(void *);
} HIST;

*/
